from __future__ import annotations

import logging
import pickle
import queue
import threading
import time
import traceback
from queue import Queue
from typing import TYPE_CHECKING

import zmq
from egse.confman import ConfigurationManagerProxy
from egse.control import Failure
from egse.control import Success
from egse.dpu.fdpu import FastCameraDPUProxy
from egse.fee.ffee import HousekeepingData
from egse.fee.ffee import aeb_state
from egse.fee.ffee import f_fee_mode
from egse.reg import RegisterMap
from egse.settings import Settings
from egse.setup import load_setup
from egse.zmq import MessageIdentifier
from textual.app import App

from .messages import AebStateChanged
from .messages import CommandFailed
from .messages import CommandThreadCrashed
from .messages import DebModeChanged
from .messages import DtcInModChanged
from .messages import ExceptionCaught
from .messages import LogRetrieved
from .messages import ObsidChanged
from .messages import OutbuffChanged
from .messages import ShutdownReached
from .messages import SyncModeChanged
from .messages import TimeoutReached

if TYPE_CHECKING:
    from .app import FastFEEApp

_LOGGER = logging.getLogger("egse.f-fee-tui")

dpu = Settings.load("DPU Processor")
cm = Settings.load("Configuration Manager Control Server")


class Command(threading.Thread):
    def __init__(self, app: App, command_q: Queue) -> None:
        super().__init__()
        self._app = app
        self._command_q = command_q
        self._f_dpu: FastCameraDPUProxy | None = None
        self._cm_cs: ConfigurationManagerProxy | None = None
        self._cancelled = threading.Event()

    def run(self) -> None:

        screen = self._app.get_screen("master")

        # This first loop is needed to make sure the Command thread is not ended when either the FastCameraDPUProxy
        # or the ConfigurationManagerProxy can not connect to their control server. When no connection, we wait for
        # 10 seconds before trying again, unless the app is terminated.

        while True:
            try:
                with FastCameraDPUProxy() as self._f_dpu, ConfigurationManagerProxy() as self._cm_cs:
                    while True:
                        if self._cancelled.is_set():
                            break
                        try:
                            target, command, args, kwargs = self._command_q.get_nowait()
                            self._command_q.task_done()
                            screen.log.info(f"Executing command '{command}'")
                            rc = self.execute_command(target, command, args, kwargs)
                            if isinstance(rc, Success):
                                msg = rc.return_code
                            else:
                                msg = rc
                            screen.log.info(f"Command '{command}' executed: {msg = !s}")
                        except queue.Empty:
                            time.sleep(0.1)  # Sleep briefly to avoid busy-waiting
                            continue
                        except Exception as exc:
                            screen.log.error(f"Caught and exception: {exc}")
                            tb = traceback.format_exc()
                            screen.post_message(ExceptionCaught(exc, tb))
            except ConnectionError as exc:
                screen.post_message(CommandThreadCrashed(exc))

            # FIXME: It would be good to know here which service is not responding, so we can still use the other
            #        proxies to send commands.

            if self.sleep_or_break(10.0):
                break

    def sleep_or_break(self, timeout: float) -> bool:
        """
        Returns True if the App is terminated and the cancelled flag is set, returns False after sleeping
        for the given timeout in seconds. The idea for this function is that when we have a sleep in the
        event loop, we still can terminate the app without having to wait the whole sleep time.
        """
        for _ in range(int(timeout * 10)):
            if self._cancelled.is_set():
                is_cancelled = True
                break
            time.sleep(0.1)
        else:
            is_cancelled = False

        return True if is_cancelled else False

    def cancel(self) -> None:
        self._cancelled.set()

    def execute_command(self, target: str, command: str, args: list, kwargs: dict):

        screen = self._app.get_screen("master")

        if target == "DPU":
            try:
                response = getattr(self._f_dpu, command)(*args, **kwargs)
                screen.log.info(f"Running DPU command: {command}({args}, {kwargs})\n"
                                f"Response: {response}")
                if isinstance(response, Failure):
                    screen.post_message(CommandFailed(response.message))
                return response
            except AttributeError as exc:
                screen.log.warning(f"No such command for DPU: {command}", exc_info=True)

        elif target == "CM_CS":
            try:
                response = getattr(self._cm_cs, command)(*args, **kwargs)
                screen.log.info(f"Running CM_CS command: {command}({args}, {kwargs})\nResponse: {response}")
                if isinstance(response, Failure):
                    screen.post_message(CommandFailed(response.message))
                return response
            except AttributeError as exc:
                screen.log.warning(f"No such command for CM_CS: {command}", exc_info=True)


class Monitor(threading.Thread):
    def __init__(self, app: App) -> None:
        self._app = app
        self._cancelled = threading.Event()
        self._reset_frame_errors = threading.Event()

        self.hostname = dpu.HOSTNAME
        self.port = dpu.DATA_DISTRIBUTION_PORT

        self.previous_deb_mode = f_fee_mode.ON_MODE
        self.previous_aeb_state = {}
        self.accumulated_outbuff = [0, 0, 0, 0, 0, 0, 0, 0]
        """The total number of OUTBUFF errors since it was last reset."""
        self.outbuff_mapping = [0, 2, 1, 3, 4, 6, 5, 7]
        """Mapping of the OUTBUFF_x and the DTC_IN_MOD, see Setup AEB_TO_T_IN_MOD"""

        super().__init__()

    def run(self) -> None:

        start_time = time.monotonic()
        """Starting time for detecting a timeout or a shutdown of the data distribution channel."""
        timeout_reported = False
        """Flag to track if a timeout has been reported or not."""
        shutdown_reported = False
        """Flag to track if a shutdown has been reported or not."""

        context = zmq.Context.instance()

        receiver = context.socket(zmq.SUB)
        receiver.setsockopt(zmq.SUBSCRIBE, MessageIdentifier.F_FEE_REGISTER_MAP.to_bytes(1, byteorder='big'))
        receiver.setsockopt(zmq.SUBSCRIBE, MessageIdentifier.SYNC_HK_DATA.to_bytes(1, byteorder='big'))
        receiver.connect(f"tcp://{self.hostname}:{self.port}")

        moni_cm: zmq.Socket = context.socket(zmq.SUB)
        moni_cm.subscribe(b'')
        moni_cm.connect(f"tcp://{cm.HOSTNAME}:{cm.MONITORING_PORT}")

        setup = load_setup()

        screen = self._app.get_screen("master")

        while True:
            if self._cancelled.is_set():
                break
            if self._reset_frame_errors.is_set():
                self._reset_frame_errors.clear()
                self.accumulated_outbuff = [0, 0, 0, 0, 0, 0, 0, 0]
                screen.post_message(OutbuffChanged(self.accumulated_outbuff))

            socket_list, _, _ = zmq.select([receiver, moni_cm], [], [], timeout=0.1)

            if receiver in socket_list:
                start_time = time.monotonic()
                timeout_reported = False
                shutdown_reported = False
                try:
                    sync_id, pickle_string = receiver.recv_multipart()
                    sync_id = int.from_bytes(sync_id, byteorder='big')
                    data = pickle.loads(pickle_string)
                    self.handle_messages(sync_id, data, setup)
                except Exception as exc:
                    tb = traceback.format_exc()
                    screen.post_message(ExceptionCaught(exc, tb))

            if moni_cm in socket_list:
                try:
                    pickle_string = moni_cm.recv()
                    data = pickle.loads(pickle_string)
                    self.handle_cm_status(data)
                except Exception as exc:
                    tb = traceback.format_exc()
                    screen.post_message(ExceptionCaught(exc, tb))

            if time.monotonic() - start_time > 6.0 and not timeout_reported:
                screen.post_message(TimeoutReached("Timeout reached after 6s on data distribution channel."))
                timeout_reported = True
            if time.monotonic() - start_time > 10.0 and not shutdown_reported:
                screen.post_message(ShutdownReached("Resetting the monitoring panels after 10s of inactivity."))
                shutdown_reported = True

        receiver.disconnect(f"tcp://{self.hostname}:{self.port}")
        receiver.close()

        screen.post_message(LogRetrieved("Monitor Thread finished ..."))
        screen.log.info("Monitor Thread finished ...")

    def cancel(self) -> None:
        self._cancelled.set()

    def reset_frame_errors(self):
        self._reset_frame_errors.set()

    def handle_cm_status(self, data: dict):
        obsid = data['obsid']
        screen = self._app.get_screen("master")
        screen.post_message(ObsidChanged(obsid))

    def handle_messages(self, sync_id, data, setup):

        screen = self._app.get_screen("master")

        if sync_id == MessageIdentifier.F_FEE_REGISTER_MAP:

            # How can we be sure the Register Map is properly synchronised with the F-FEE?

            register_map, _ = data
            register_map = RegisterMap("F-FEE", memory_map=register_map, setup=setup)

            t0 = register_map["DEB_DTC_IN_MOD_2", "T0_IN_MOD"]
            t1 = register_map["DEB_DTC_IN_MOD_2", "T1_IN_MOD"]
            t2 = register_map["DEB_DTC_IN_MOD_2", "T2_IN_MOD"]
            t3 = register_map["DEB_DTC_IN_MOD_2", "T3_IN_MOD"]
            t4 = register_map["DEB_DTC_IN_MOD_1", "T4_IN_MOD"]
            t5 = register_map["DEB_DTC_IN_MOD_1", "T5_IN_MOD"]
            t6 = register_map["DEB_DTC_IN_MOD_1", "T6_IN_MOD"]
            t7 = register_map["DEB_DTC_IN_MOD_1", "T7_IN_MOD"]

            screen.log(f"{t0=}, {t1=}, {t2=}, {t3=}, {t4=}, {t5=}, {t6=}, {t7=}")

            screen.post_message(DtcInModChanged(t0, t1, t2, t3, t4, t5, t6, t7))

            sync_mode = register_map["DEB_DTC_SEL_TRG", "TRG_SRC"]

            screen.post_message(SyncModeChanged(sync_mode))

        elif sync_id == MessageIdentifier.SYNC_HK_DATA:

            cmd, aeb_id, data, timestamp = data

            if cmd == 'command_deb_read_hk':
                hk_data = HousekeepingData("DEB", data, setup)
                deb_mode = f_fee_mode(hk_data["STATUS", "OPER_MOD"])
                screen.log(f"DEB_MODE = {deb_mode.name}")
                screen.post_message(DebModeChanged(deb_mode))
                outbuff = [hk_data["OVF", f"OUTBUFF_{x}"] for x in range(1, 9)]
                # outbuff = random.choices([0, 1], k=8)
                screen.log(f"{outbuff = }")
                if any(outbuff):
                    # re-order outbuff to match with DTC_IN_MOD
                    outbuff = [outbuff[idx] for idx in self.outbuff_mapping]
                    # update the accumulated values for OUTBUFF
                    self.accumulated_outbuff = [x + y for x, y in zip(self.accumulated_outbuff, outbuff)]
                    screen.post_message(OutbuffChanged(self.accumulated_outbuff))
            elif cmd == 'command_aeb_read_hk':
                aeb_id = aeb_id[0]  # this comes from the args, so it's a list
                hk_data = HousekeepingData(aeb_id, data, setup)
                aeb_status = hk_data["STATUS", "AEB_STATUS"]
                screen.log(f"AEB_STATE = {aeb_state(aeb_status).name}")

                # Power is/was handled by the DEB_DTC_AEB_ONOFF state from the DEB register map above.
                # Should I check the AEB_ONOFF from the Register Map? Is there a chance/reason why the AEB_STATE
                # would not match?

                if aeb_status == aeb_state.OFF:
                    screen.post_message(AebStateChanged(f"{aeb_id.lower()}_onoff", False))
                elif aeb_status == aeb_state.INIT:
                    screen.post_message(AebStateChanged(f"{aeb_id.lower()}_init", True))
                elif aeb_status == aeb_state.POWER_UP:
                    screen.post_message(AebStateChanged(f"{aeb_id.lower()}_power_up", True))
                elif aeb_status == aeb_state.POWER_DOWN:
                    screen.post_message(AebStateChanged(f"{aeb_id.lower()}_power_down", True))
                elif aeb_status == aeb_state.CONFIG:
                    screen.post_message(AebStateChanged(f"{aeb_id.lower()}_config", True))
                elif aeb_status == aeb_state.IMAGE:
                    screen.post_message(AebStateChanged(f"{aeb_id.lower()}_image", True))
                elif aeb_status == aeb_state.PATTERN:
                    screen.post_message(AebStateChanged(f"{aeb_id.lower()}_pattern", True))



from __future__ import annotations

import ast
import itertools
from typing import Tuple

from textual.app import ComposeResult
from textual.widgets import Label
from textual.widgets import Sparkline
from textual.widgets import Static

from f_fee_tui.messages import DtcInModChanged
from egse.setup import load_setup

setup = load_setup()

ON = "✖"
OFF = ""

# From Setup: setup.camera.fee.ccd_numbering.AEB_TO_T_IN_MOD

# │   │   │   └── AEB_TO_T_IN_MOD
# │   │   │       ├── (1, F): ((0b001, 0, 0, 0), (0, 0, 0, 0))
# │   │   │       ├── (1, E): ((0, 0, 0b010, 0), (0, 0, 0, 0))
# │   │   │       ├── (2, F): ((0, 0b010, 0, 0), (0, 0, 0, 0))
# │   │   │       ├── (2, E): ((0, 0, 0, 0b001), (0, 0, 0, 0))
# │   │   │       ├── (3, F): ((0, 0, 0, 0), (0b001, 0, 0, 0))
# │   │   │       ├── (3, E): ((0, 0, 0, 0), (0, 0, 0b010, 0))
# │   │   │       ├── (4, F): ((0, 0, 0, 0), (0, 0b010, 0, 0))
# │   │   │       └── (4, E): ((0, 0, 0, 0), (0, 0, 0, 0b001))

AEB_TO_T_IN_MOD = setup.camera.fee.ccd_numbering.AEB_TO_T_IN_MOD

# This translation table is used to map the widget ID to the value of a Tx_IN_MOD.

T0 = {0b001: "T0-001"}                   # AEB1-F
T1 = {0b001: "T1-001", 0b010: "T1-010"}  # AEB1-E & AEB2-F
T2 = {0b001: "T2-001", 0b010: "T2-010"}  # AEB2-F & AEB1-E
T3 = {0b001: "T3-001"}                   # AEB2-E
T4 = {0b001: "T4-001"}                   # AEB3-F
T5 = {0b001: "T5-001", 0b010: "T5-010"}  # AEB3-E & AEB4-F
T6 = {0b001: "T6-001", 0b010: "T6-010"}  # AEB4-F & AEB3-E
T7 = {0b001: "T7-001"}                   # AEB4-E

# The following list is used below to clear all the widgets with that label id.

ALL_IDS = list(itertools.chain(
        T0.values(), T1.values(), T2.values(), T3.values(), T4.values(), T5.values(), T6.values(), T7.values()
))


# Map AEB id and side to a Tx_IN_MOD

def aeb_to_tx_in_mod(aeb_id: str, side: str) -> Tuple[int, int] | None:
    """For a given AEB number and side, this function extracts the index and value of the Tx_IN_MOD."""
    aeb_nr = aeb_id[-1]
    key = f"({aeb_nr}, {side})"

    values = list(itertools.chain.from_iterable(ast.literal_eval(AEB_TO_T_IN_MOD[key])))

    for idx, value in enumerate(values):
        if value:
            return idx, value
    else:
        return None


def tx_in_md_to_aeb(in_mod: int) -> Tuple[str, str]:
    ...


class DtcInMod(Static):
    """A widget to monitor the state of the AEBs."""

    def compose(self) -> ComposeResult:
        yield Label("", classes="header")
        yield Label("AEB1", classes="two-cols header")
        yield Label("AEB2", classes="two-cols header")
        yield Label("AEB3", classes="two-cols header")
        yield Label("AEB4", classes="two-cols header")

        yield Label("SpW 1")
        yield Label("", id="T0-001", classes="one-col")    # AEB1-F
        yield Label("", id="T1-001", classes="one-col")    # AEB1-E
        yield Label("", id="T1-010", classes="one-col")    # AEB2-F
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")

        yield Label("SpW 2")
        yield Label("", classes="one-col disabled")
        yield Label("", id="T2-010", classes="one-col")    # AEB1-E
        yield Label("", id="T2-001", classes="one-col")    # AEB2-F
        yield Label("", id="T3-001", classes="one-col")    # AEB2-E
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")

        yield Label("SpW 3")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", id="T4-001", classes="one-col")    # AEB3-F
        yield Label("", id="T5-001", classes="one-col")    # AEB3-E
        yield Label("", id="T5-010", classes="one-col")    # AEB4-F
        yield Label("", classes="one-col disabled")

        yield Label("SpW 4")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", id="T6-010", classes="one-col")    # AEB3-E
        yield Label("", id="T6-001", classes="one-col")    # AEB4-F
        yield Label("", id="T7-001", classes="one-col")    # AEB4-E

        yield Label("", classes="footer")
        yield Label("F", classes="one-col footer")
        yield Label("E", classes="one-col footer")
        yield Label("F", classes="one-col footer")
        yield Label("E", classes="one-col footer")
        yield Label("F", classes="one-col footer")
        yield Label("E", classes="one-col footer")
        yield Label("F", classes="one-col footer")
        yield Label("E", classes="one-col footer")

        yield Label("Errors", id="lbl-frame-errors")
        yield Sparkline([0, 0, 0, 0, 0, 0, 0, 0], id="frame-errors")

    def on_mount(self) -> None:
        tooltip_msg = "Accumulated errors on SpW transmit buffer overflow."
        self.query_one("#lbl-frame-errors", Label).tooltip = tooltip_msg
        self.query_one("#frame-errors", Sparkline).tooltip = tooltip_msg

    def clear(self):
        for id_ in ALL_IDS:
            self.query_one(f"#{id_}", Label).update(OFF)

    def set_state(self, state: DtcInModChanged):
        self.log(f"{state.t0=}, {state.t1=}, {state.t2=}, {state.t3=}, {state.t4=}, {state.t5=}, {state.t6=}, {state.t7=}")

        self.clear()

        if state.t0 == 0b001:
            self.query_one("#T0-001", Label).update(ON)

        if state.t1 == 0b001:
            self.query_one("#T1-001", Label).update(ON)
        elif state.t1 == 0b010:
            self.query_one("#T1-010", Label).update(ON)

        if state.t2 == 0b001:
            self.query_one("#T2-001", Label).update(ON)
        elif state.t2 == 0b010:
            self.query_one("#T2-010", Label).update(ON)

        if state.t3 == 0b001:
            self.query_one("#T3-001", Label).update(ON)

        if state.t4 == 0b001:
            self.query_one("#T4-001", Label).update(ON)

        if state.t5 == 0b001:
            self.query_one("#T5-001", Label).update(ON)
        elif state.t5 == 0b010:
            self.query_one("#T5-010", Label).update(ON)

        if state.t6 == 0b001:
            self.query_one("#T6-001", Label).update(ON)
        elif state.t6 == 0b010:
            self.query_one("#T6-010", Label).update(ON)

        if state.t7 == 0b001:
            self.query_one("#T7-001", Label).update(ON)



from queue import Queue
from queue import Empty


class ClearableQueue(Queue):
    """Add the possibility to clear the queue."""
    def clear(self):
        msg = []
        try:
            while True:
                msg.append(str(self.get_nowait()))
                self.task_done()
        except Empty:
            pass
        finally:
            if msg:
                print("Queue not empty at exit:")
                print("\n".join(msg))



from __future__ import annotations

from typing import Optional

from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal
from textual.containers import Vertical
from textual.screen import ModalScreen
from textual.widgets import Button
from textual.widgets import Input
from textual.widgets import Label
from textual.widgets import Markdown
from textual.widgets import Select


class InputStringDialog(ModalScreen[Optional[str]]):
    """A modal dialog asking for an input string."""

    BINDINGS = [Binding("escape", "dismiss(False)", "", show=False)]

    def __init__(self, default_input: str, title: str = None):
        super().__init__()

        self._title = title or "Question"
        self._default_input = default_input
        self._description = default_input

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label(self._title)
            yield Input(placeholder=self._default_input)
            with Horizontal():
                yield Button("OK", variant="primary", id="btn-dialog-ok")
                yield Button("Cancel", id="btn-dialog-cancel")

    @on(Input.Changed)
    def save_description(self, event: Input.Changed) -> None:
        self._description = event.value

    @on(Input.Submitted)
    def return_pressed(self, event: Input.Submitted):
        self._description = event.value

    @on(Button.Pressed)
    def on_button_pressed(self, event: Button.Pressed) -> None:
        """React to button press."""

        if event.button.id == "btn-dialog-ok":
            self.dismiss(self._description or self._default_input)
            # self.app.query_one("#console-log", ConsoleOutput).write_log_info("You pressed YES...")
        else:
            self.dismiss(None)
            # self.app.query_one("#console-log", ConsoleOutput).write_log_info("You pressed NO...")


class YesNoDialog(ModalScreen[bool]):
    """A modal dialog that asks a YES / NO question."""

    BINDINGS = [Binding("escape", "dismiss(False)", "", show=False)]

    def __init__(self, question: str, title: str = None):
        super().__init__()

        self._title = title or "Question"
        self._question = question

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label(self._title)
            yield Markdown(self._question)
            with Horizontal():
                yield Button("Yes", variant="primary", id="btn-dialog-yes")
                yield Button("No", id="btn-dialog-no")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """React to button press."""

        if event.button.id == "btn-dialog-yes":
            self.dismiss(True)
            # self.app.query_one("#console-log", ConsoleOutput).write_log_info("You pressed YES...")
        else:
            self.dismiss(False)
            # self.app.query_one("#console-log", ConsoleOutput).write_log_info("You pressed NO...")


class SelectDialog(ModalScreen[int]):
    """A modal dialog with a selection from which to choose."""

    BINDINGS = [Binding("escape", "dismiss(-1)", "", show=False)]

    def __init__(self, options: list[str], title: str = None, message: str = None, footer: str = None):
        super().__init__()

        self._title = title or "Select one of the options"
        self._message = message or ""
        self._options = [(str(y), x) for x, y in enumerate(options)]
        self._footer = footer or "Press [i]escape[/] to dismiss this dialog."
        self._selected_index = -1

    def compose(self) -> ComposeResult:

        with Vertical():
            yield Label(self._title)
            yield Markdown(self._message)
            yield Select(self._options)
            with Horizontal():
                yield Button("OK", variant="primary", id="btn-dialog-ok")
                yield Button("Cancel", id="btn-dialog-cancel")
            if self._footer:
                yield Label(self._footer, id="lbl-dialog-footer")

    @on(Select.Changed)
    def select_changed(self, event: Select.Changed) -> None:
        self._selected_index = event.value

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """React to button press."""

        if event.button.id == "btn-dialog-ok":
            self.dismiss(self._selected_index)
        else:
            self.dismiss(-1)



"""
This gist is a stand alone Textual App intended to show how to handle Threading in a Textual App, especially when
the code in the thread `run()` method would crash or raise an Exception. It also demonstrates how messages shall be
posted to either the App or the Screen depending on where the receiving method is implemented.

There are two Messages, `RuntimeErrorCaught` and `ThreadCrashed`. Both messages are sent from the Thread, the former
will be sent to the App, the latter needs to be sent to the 'master' Screen. That is because the receiving methods are
implemented in the App and in the 'master' screen respectively.

This gist is mainly written as a reminder to myself for future use, but I hope it will be useful to other Textual users.
"""
import datetime
import random
import threading
import time
from queue import Queue

from textual.app import App
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Vertical
from textual.message import Message
from textual.screen import Screen
from textual.widgets import Footer
from textual.widgets import Header
from textual.widgets import RichLog
from textual.widgets import Static


class RuntimeErrorCaught(Message):
    """A Message that is sent to the App."""
    def __init__(self, message: str, exc: Exception):
        super().__init__()
        self.msg = message
        self.exc = exc


class ThreadCrashed(Message):
    """A Message that is sent to the MasterScreen."""
    def __init__(self, message: str, exc: Exception):
        super().__init__()
        self.msg = message
        self.exc = exc


class Command(threading.Thread):
    """
    A thread that is used to execute commands on services in the background and report back to the App with
    posted messages. Commands are passed using the thread-safe Queue.
    """
    def __init__(self, app: App, command_q: Queue):
        super().__init__()
        self._app = app
        self._command_q = command_q
        self._cancelled = threading.Event()

    def run(self):
        self._app.log("Command thread started ...")

        while True:
            if self._cancelled.is_set():
                break

            try:
                # Within this try ... except: clause, you would execute your code which might connect to some
                # services, execute complex commands it gets from the command queue and might raise an exception.
                # The code snippet below simulates an Exception that is thrown after some time.

                if random.random() < 0.1:
                    raise RuntimeError("A fake runtime error.")

            except RuntimeError as exc:
                msg = "We got a RuntimeError in the Command thread ..."
                self._app.get_screen("master").post_message(ThreadCrashed(msg, exc))
                self._app.post_message(RuntimeErrorCaught(msg, exc))

                if self.sleep_or_break():
                    break

                self._app.notify("Re-activating Command Thread after 10.0s")

    def sleep_or_break(self) -> bool:
        """
        Sleep for 10 seconds. If the thread was cancelled, break early and return True.
        """
        for _ in range(100):
            if self._cancelled.is_set():
                is_cancelled = True
                break
            time.sleep(0.1)
        else:
            is_cancelled = False

        return True if is_cancelled else False

    def cancel(self) -> None:
        self._cancelled.set()


class MasterScreen(Screen):

    DEFAULT_CSS = """
    #lbl-msg {
        width: auto;
        padding: 1 2;
        border: wide white;
    }
    """

    def __init__(self):
        super().__init__()
        self._command_q = Queue()
        self._commanding_thread = Command(self.app, self._command_q)

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header()
        yield Footer()
        with Vertical():
            yield Static("Gist for demonstrating how threading works with Textual.", id="lbl-msg")
            yield RichLog(max_lines=200, markup=True, id="rich-log")

    def on_mount(self) -> None:
        self._commanding_thread.start()

    def on_unmount(self) -> None:
        self._commanding_thread.cancel()

        if self._commanding_thread.is_alive():
            self._commanding_thread.join()

    def on_thread_crashed(self, message: ThreadCrashed):
        self.log(f"MasterScreen: {message.msg}: {message.exc}")
        self.query_one("#rich-log", RichLog).write(f"{datetime.datetime.now():%X} [green]MasterScreen:[/] {message.msg} — {message.exc}")


class ThreadApp(App):
    BINDINGS = [
        Binding("q", "quit", "Quit"),
    ]

    SCREENS = {"master": MasterScreen}

    def on_mount(self):
        self.push_screen("master")

    def on_runtime_error_caught(self, message: RuntimeErrorCaught):
        self.log(f"App: {message.msg}: {message.exc}")
        self.query_one("#rich-log", RichLog).write(f"{datetime.datetime.now():%X} [red]App:[/] {message.msg} — {message.exc}")


if __name__ == '__main__':

    app = ThreadApp()
    app.run()
